from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from dagster import check
from dagster.config.config_type import ALL_CONFIG_BUILTINS
from dagster.utils import file_relative_path

from .graphql_context_test_suite import ReadonlyGraphQLContextTestMatrix
from .setup import csv_hello_world_solids_config

CONFIG_VALIDATION_QUERY = '''
query PipelineQuery(
    $runConfigData: RunConfigData,
    $pipeline: PipelineSelector!,
    $mode: String!
) {
    isPipelineConfigValid(
        runConfigData: $runConfigData,
        pipeline: $pipeline,
        mode: $mode
    ) {
        __typename
        ... on PipelineConfigValidationValid {
            pipelineName
        }
        ... on PipelineConfigValidationInvalid {
            pipelineName
            errors {
                __typename
                ... on RuntimeMismatchConfigError {
                    valueRep
                }
                ... on MissingFieldConfigError {
                    field { name }
                }
                ... on MissingFieldsConfigError {
                    fields { name }
                }
                ... on FieldNotDefinedConfigError {
                    fieldName
                }
                ... on FieldsNotDefinedConfigError {
                    fieldNames
                }
                ... on SelectorTypeConfigError {
                    incomingFields
                }
                message
                reason
                stack {
                    entries {
                        __typename
                        ... on EvaluationStackPathEntry {
                            fieldName
                        }
                        ... on EvaluationStackListItemEntry {
                            listIndex
                        }
                    }
                }
            }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
        ... on PythonError {
            message
            stack
        }
    }
}
'''


def field_stack(error_data):
    return [
        entry['fieldName']
        for entry in error_data['stack']['entries']
        if entry['__typename'] == 'EvaluationStackPathEntry'
    ]


def single_error_data(result):
    assert len(result.data['isPipelineConfigValid']['errors']) == 1
    return result.data['isPipelineConfigValid']['errors'][0]


def find_error(result, field_stack_to_find, reason):
    llist = list(find_errors(result, field_stack_to_find, reason))
    assert len(llist) == 1
    return llist[0]


def find_errors(result, field_stack_to_find, reason):
    error_datas = result.data['isPipelineConfigValid']['errors']
    for error_data in error_datas:
        if field_stack_to_find == field_stack(error_data) and error_data['reason'] == reason:
            yield error_data


def execute_config_graphql(context, pipeline_name, run_config, mode):
    selector = infer_pipeline_selector(context, pipeline_name)
    return execute_dagster_graphql(
        context,
        CONFIG_VALIDATION_QUERY,
        {'runConfigData': run_config, 'pipeline': selector, 'mode': mode,},
    )


class TestConfigTypes(ReadonlyGraphQLContextTestMatrix):
    def test_pipeline_not_found(self, graphql_context):
        result = execute_config_graphql(
            graphql_context, pipeline_name='nope', run_config={}, mode='default'
        )

        assert not result.errors
        assert result.data
        assert result.data['isPipelineConfigValid']['__typename'] == 'PipelineNotFoundError'
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'nope'

    def test_basic_valid_config(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='csv_hello_world',
            run_config=csv_hello_world_solids_config(),
            mode='default',
        )

        assert not result.errors
        assert result.data
        assert result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationValid'
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'csv_hello_world'

    def test_root_field_not_defined(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='csv_hello_world',
            run_config={
                'solids': {
                    'sum_solid': {
                        'inputs': {'num': file_relative_path(__file__, '../data/num.csv')}
                    }
                },
                'nope': {},
            },
            mode='default',
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'csv_hello_world'
        errors = result.data['isPipelineConfigValid']['errors']
        assert len(errors) == 1
        error = errors[0]
        assert error['__typename'] == 'FieldNotDefinedConfigError'
        assert error['fieldName'] == 'nope'
        assert not error['stack']['entries']

    def test_basic_invalid_not_defined_field(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='csv_hello_world',
            run_config={'solids': {'sum_solid': {'inputs': {'num': 'foo.txt', 'extra': 'nope'}}}},
            mode='default',
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'csv_hello_world'
        assert len(result.data['isPipelineConfigValid']['errors']) == 1
        error_data = result.data['isPipelineConfigValid']['errors'][0]
        assert ['solids', 'sum_solid', 'inputs'] == field_stack(error_data)
        assert error_data['reason'] == 'FIELD_NOT_DEFINED'
        assert error_data['fieldName'] == 'extra'

    def test_multiple_not_defined_fields(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='csv_hello_world',
            run_config={
                'solids': {
                    'sum_solid': {
                        'inputs': {'num': 'foo.txt', 'extra_one': 'nope', 'extra_two': 'nope'}
                    }
                }
            },
            mode='default',
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'csv_hello_world'
        assert len(result.data['isPipelineConfigValid']['errors']) == 1
        error_data = result.data['isPipelineConfigValid']['errors'][0]
        assert ['solids', 'sum_solid', 'inputs'] == field_stack(error_data)
        assert error_data['reason'] == 'FIELDS_NOT_DEFINED'
        assert error_data['fieldNames'] == ['extra_one', 'extra_two']

    def test_root_wrong_type(self, graphql_context):
        result = execute_config_graphql(
            graphql_context, pipeline_name='csv_hello_world', run_config=123, mode='default'
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'csv_hello_world'
        assert len(result.data['isPipelineConfigValid']['errors']) == 1
        error_data = result.data['isPipelineConfigValid']['errors'][0]
        assert error_data['reason'] == 'RUNTIME_TYPE_MISMATCH'

    def test_basic_invalid_config_type_mismatch(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='csv_hello_world',
            run_config={'solids': {'sum_solid': {'inputs': {'num': 123}}}},
            mode='default',
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'csv_hello_world'
        assert len(result.data['isPipelineConfigValid']['errors']) == 1
        error_data = result.data['isPipelineConfigValid']['errors'][0]
        assert error_data['message']
        assert error_data['stack']
        assert error_data['stack']['entries']
        assert error_data['reason'] == 'RUNTIME_TYPE_MISMATCH'
        assert error_data['valueRep'] == '123'

        assert ['solids', 'sum_solid', 'inputs', 'num'] == field_stack(error_data)

    def test_basic_invalid_config_missing_field(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='csv_hello_world',
            run_config={'solids': {'sum_solid': {'inputs': {}}}},
            mode='default',
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'csv_hello_world'
        assert len(result.data['isPipelineConfigValid']['errors']) == 1
        error_data = result.data['isPipelineConfigValid']['errors'][0]

        assert ['solids', 'sum_solid', 'inputs'] == field_stack(error_data)
        assert error_data['reason'] == 'MISSING_REQUIRED_FIELD'
        assert error_data['field']['name'] == 'num'

    def test_mode_resource_config_works(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='multi_mode_with_resources',
            run_config={'resources': {'op': {'config': 2}}},
            mode='add_mode',
        )

        assert not result.errors
        assert result.data
        assert result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationValid'
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'multi_mode_with_resources'

        result = execute_config_graphql(
            graphql_context,
            pipeline_name='multi_mode_with_resources',
            run_config={'resources': {'op': {'config': 2}}},
            mode='mult_mode',
        )

        assert not result.errors
        assert result.data
        assert result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationValid'
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'multi_mode_with_resources'

        result = execute_config_graphql(
            graphql_context,
            pipeline_name='multi_mode_with_resources',
            run_config={'resources': {'op': {'config': {'num_one': 2, 'num_two': 3}}}},
            mode='double_adder',
        )

        assert not result.errors
        assert result.data
        assert result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationValid'
        assert result.data['isPipelineConfigValid']['pipelineName'] == 'multi_mode_with_resources'

    def test_missing_resource(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='multi_mode_with_resources',
            run_config={'resources': {}},
            mode='add_mode',
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        error_data = single_error_data(result)
        assert error_data['reason'] == 'MISSING_REQUIRED_FIELD'
        assert error_data['field']['name'] == 'op'

    def test_undefined_resource(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='multi_mode_with_resources',
            run_config={'resources': {'nope': {}}},
            mode='add_mode',
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['isPipelineConfigValid']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert {'FieldNotDefinedConfigError', 'MissingFieldConfigError'} == {
            error_data['__typename']
            for error_data in result.data['isPipelineConfigValid']['errors']
        }

    def test_more_complicated_works(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='more_complicated_nested_config',
            run_config={
                'solids': {
                    'a_solid_with_multilayered_config': {
                        'config': {
                            'field_any': {'123': 123},
                            'field_one': 'foo.txt',
                            'field_two': 'yup',
                            'field_three': 'mmmhmmm',
                            'nested_field': {'field_four_str': 'yaya', 'field_five_int': 234},
                        }
                    }
                }
            },
            mode='default',
        )
        assert not result.errors
        assert result.data
        valid_data = result.data['isPipelineConfigValid']
        assert valid_data['__typename'] == 'PipelineConfigValidationValid'
        assert valid_data['pipelineName'] == 'more_complicated_nested_config'

    def test_multiple_missing_fields(self, graphql_context):

        result = execute_config_graphql(
            graphql_context,
            pipeline_name='more_complicated_nested_config',
            run_config={'solids': {'a_solid_with_multilayered_config': {'config': {}}}},
            mode='default',
        )

        assert not result.errors
        assert result.data
        valid_data = result.data['isPipelineConfigValid']

        assert valid_data['__typename'] == 'PipelineConfigValidationInvalid'
        assert valid_data['pipelineName'] == 'more_complicated_nested_config'
        assert len(valid_data['errors']) == 1
        error_data = valid_data['errors'][0]
        missing_names = {field_data['name'] for field_data in error_data['fields']}
        assert missing_names == {'nested_field', 'field_one', 'field_any'}
        assert field_stack(error_data) == ['solids', 'a_solid_with_multilayered_config', 'config']

    def test_more_complicated_multiple_errors(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='more_complicated_nested_config',
            run_config={
                'solids': {
                    'a_solid_with_multilayered_config': {
                        'config': {
                            'field_any': [],
                            # 'field_one': 'foo.txt', # missing
                            'field_two': 'yup',
                            'field_three': 'mmmhmmm',
                            'extra_one': 'kjsdkfjd',  # extra
                            'nested_field': {
                                'field_four_str': 23434,  # runtime type
                                'field_five_int': 234,
                                'extra_two': 'ksjdkfjd',  # another extra
                            },
                        }
                    }
                }
            },
            mode='default',
        )

        assert not result.errors
        assert result.data
        valid_data = result.data['isPipelineConfigValid']

        assert valid_data['__typename'] == 'PipelineConfigValidationInvalid'
        assert valid_data['pipelineName'] == 'more_complicated_nested_config'
        assert len(valid_data['errors']) == 4

        missing_error_one = find_error(
            result,
            ['solids', 'a_solid_with_multilayered_config', 'config'],
            'MISSING_REQUIRED_FIELD',
        )
        assert ['solids', 'a_solid_with_multilayered_config', 'config'] == field_stack(
            missing_error_one
        )
        assert missing_error_one['reason'] == 'MISSING_REQUIRED_FIELD'
        assert missing_error_one['field']['name'] == 'field_one'

        not_defined_one = find_error(
            result, ['solids', 'a_solid_with_multilayered_config', 'config'], 'FIELD_NOT_DEFINED'
        )
        assert ['solids', 'a_solid_with_multilayered_config', 'config'] == field_stack(
            not_defined_one
        )
        assert not_defined_one['reason'] == 'FIELD_NOT_DEFINED'
        assert not_defined_one['fieldName'] == 'extra_one'

        dagster_type_error = find_error(
            result,
            [
                'solids',
                'a_solid_with_multilayered_config',
                'config',
                'nested_field',
                'field_four_str',
            ],
            'RUNTIME_TYPE_MISMATCH',
        )
        assert [
            'solids',
            'a_solid_with_multilayered_config',
            'config',
            'nested_field',
            'field_four_str',
        ] == field_stack(dagster_type_error)
        assert dagster_type_error['reason'] == 'RUNTIME_TYPE_MISMATCH'
        assert dagster_type_error['valueRep'] == '23434'

        not_defined_two = find_error(
            result,
            ['solids', 'a_solid_with_multilayered_config', 'config', 'nested_field'],
            'FIELD_NOT_DEFINED',
        )

        assert [
            'solids',
            'a_solid_with_multilayered_config',
            'config',
            'nested_field',
        ] == field_stack(not_defined_two)
        assert not_defined_two['reason'] == 'FIELD_NOT_DEFINED'
        assert not_defined_two['fieldName'] == 'extra_two'

        # TODO: two more errors

    def test_config_list(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='pipeline_with_list',
            run_config={'solids': {'solid_with_list': {'config': [1, 2]}}},
            mode='default',
        )

        assert not result.errors
        assert result.data
        valid_data = result.data['isPipelineConfigValid']
        assert valid_data['__typename'] == 'PipelineConfigValidationValid'
        assert valid_data['pipelineName'] == 'pipeline_with_list'

    def test_config_list_invalid(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='pipeline_with_list',
            run_config={'solids': {'solid_with_list': {'config': 'foo'}}},
            mode='default',
        )

        assert not result.errors
        assert result.data
        valid_data = result.data['isPipelineConfigValid']
        assert valid_data['__typename'] == 'PipelineConfigValidationInvalid'
        assert valid_data['pipelineName'] == 'pipeline_with_list'
        assert len(valid_data['errors']) == 1
        assert ['solids', 'solid_with_list', 'config'] == field_stack(valid_data['errors'][0])

    def test_config_list_item_invalid(self, graphql_context):
        result = execute_config_graphql(
            graphql_context,
            pipeline_name='pipeline_with_list',
            run_config={'solids': {'solid_with_list': {'config': [1, 'foo']}}},
            mode='default',
        )

        assert not result.errors
        assert result.data
        valid_data = result.data['isPipelineConfigValid']
        assert valid_data['__typename'] == 'PipelineConfigValidationInvalid'
        assert valid_data['pipelineName'] == 'pipeline_with_list'
        assert len(valid_data['errors']) == 1
        entries = valid_data['errors'][0]['stack']['entries']
        assert len(entries) == 4
        assert ['solids', 'solid_with_list', 'config'] == field_stack(valid_data['errors'][0])

        last_entry = entries[3]
        assert last_entry['__typename'] == 'EvaluationStackListItemEntry'
        assert last_entry['listIndex'] == 1

    def test_smoke_test_config_type_system(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'more_complicated_nested_config')
        result = execute_dagster_graphql(
            graphql_context, ALL_CONFIG_TYPES_QUERY, {'selector': selector, 'mode': 'default'},
        )

        config_types_data = result.data['runConfigSchemaOrError']['allConfigTypes']

        assert has_config_type_with_key_prefix(config_types_data, 'Shape.')

        for builtin_config_type in ALL_CONFIG_BUILTINS:
            assert has_config_type(config_types_data, builtin_config_type.given_name)


def pipeline_named(result, name):
    for pipeline_data in result.data['pipelines']['nodes']:
        if pipeline_data['name'] == name:
            return pipeline_data
    check.failed('Did not find')


def has_config_type_with_key_prefix(config_types_data, prefix):
    for config_type_data in config_types_data:
        if config_type_data['key'].startswith(prefix):
            return True

    return False


def has_config_type(config_types_data, name):
    for config_type_data in config_types_data:
        if config_type_data.get('givenName') == name:
            return True

    return False


ALL_CONFIG_TYPES_QUERY = '''
fragment configTypeFragment on ConfigType {
  __typename
  key
  description
  isSelector
  typeParamKeys
  recursiveConfigTypes {
    key
    description
    ... on CompositeConfigType {
        fields {
            name
            isRequired
            description
        }
    }
    ... on WrappingConfigType {
        ofType { key }
    }
  }
  ... on EnumConfigType {
    givenName
    values {
      value
      description
    }
  }
  ... on RegularConfigType {
    givenName
  }
  ... on CompositeConfigType {
    fields {
      name
      isRequired
      description
    }
  }
  ... on WrappingConfigType {
    ofType { key }
  }
  ... on ScalarUnionConfigType {
    scalarType { key }
    nonScalarType { key }
  }
}

query allConfigTypes($selector: PipelineSelector!, $mode: String!) {
  runConfigSchemaOrError(selector: $selector, mode: $mode ) {
    ... on RunConfigSchema {
      allConfigTypes {
        ...configTypeFragment
      }
    }
  }
}
'''


def get_field_data(config_type_data, name):
    for field_data in config_type_data['fields']:
        if field_data['name'] == name:
            return field_data


def get_field_names(config_type_data):
    return {field_data['name'] for field_data in config_type_data.get('fields', [])}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestMaterializations.test_materializations[in_memory_instance_in_process_env] 1'] = [
    'PipelineStartEvent',
    'ExecutionStepStartEvent',
    'StepMaterializationEvent',
    'ExecutionStepOutputEvent',
    'ExecutionStepSuccessEvent',
    'PipelineSuccessEvent'
]

snapshots['TestMaterializations.test_materializations[sqlite_with_sync_run_launcher_in_process_env] 1'] = [
    'PipelineStartEvent',
    'ExecutionStepStartEvent',
    'StepMaterializationEvent',
    'ExecutionStepOutputEvent',
    'ExecutionStepSuccessEvent',
    'PipelineSuccessEvent'
]

snapshots['TestMaterializations.test_materializations[sqlite_with_cli_api_run_launcher_in_process_env] 1'] = [
    'PipelineStartEvent',
    'ExecutionStepStartEvent',
    'StepMaterializationEvent',
    'ExecutionStepOutputEvent',
    'ExecutionStepSuccessEvent',
    'PipelineSuccessEvent'
]

snapshots['TestMaterializations.test_materializations[sqlite_with_grpc_run_launcher_in_process_env] 1'] = [
    'PipelineStartEvent',
    'ExecutionStepStartEvent',
    'StepMaterializationEvent',
    'ExecutionStepOutputEvent',
    'ExecutionStepSuccessEvent',
    'PipelineSuccessEvent'
]



from dagster_graphql.client.query import (
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
    LAUNCH_PIPELINE_REEXECUTION_MUTATION,
)
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from dagster.core.utils import make_new_run_id

from .graphql_context_test_suite import ExecutingGraphQLContextTestMatrix
from .setup import csv_hello_world_solids_config, csv_hello_world_solids_config_fs_storage

RUN_QUERY = '''
query RunQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on PipelineRun {
        status
      }
    }
  }
'''


class TestReexecution(ExecutingGraphQLContextTestMatrix):
    def test_full_pipeline_reexecution_fs_storage(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        run_id = make_new_run_id()
        result_one = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config_fs_storage(),
                    'executionMetadata': {'runId': run_id},
                    'mode': 'default',
                }
            },
        )

        assert (
            result_one.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        )

        result_one.data['launchPipelineExecution']['run']['runId'] = '<runId dummy value>'
        result_one.data['launchPipelineExecution']['run'][
            'runConfigYaml'
        ] = '<runConfigYaml dummy value>'

        snapshot.assert_match(result_one.data)

        # reexecution
        new_run_id = make_new_run_id()

        result_two = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config_fs_storage(),
                    'executionMetadata': {
                        'runId': new_run_id,
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                    },
                    'mode': 'default',
                }
            },
        )

        query_result = result_two.data['launchPipelineReexecution']
        assert query_result['__typename'] == 'LaunchPipelineRunSuccess'
        assert query_result['run']['rootRunId'] == run_id
        assert query_result['run']['parentRunId'] == run_id

    def test_full_pipeline_reexecution_in_memory_storage(self, graphql_context, snapshot):
        run_id = make_new_run_id()
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result_one = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config(),
                    'executionMetadata': {'runId': run_id},
                    'mode': 'default',
                }
            },
        )

        assert (
            result_one.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        )

        result_one.data['launchPipelineExecution']['run']['runId'] = '<runId dummy value>'
        result_one.data['launchPipelineExecution']['run'][
            'runConfigYaml'
        ] = '<runConfigYaml dummy value>'

        snapshot.assert_match(result_one.data)

        # reexecution
        new_run_id = make_new_run_id()

        result_two = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config(),
                    'executionMetadata': {
                        'runId': new_run_id,
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                    },
                    'mode': 'default',
                }
            },
        )

        query_result = result_two.data['launchPipelineReexecution']
        assert query_result['__typename'] == 'LaunchPipelineRunSuccess'
        assert query_result['run']['rootRunId'] == run_id
        assert query_result['run']['parentRunId'] == run_id

    def test_pipeline_reexecution_successful_launch(self, graphql_context):
        instance = graphql_context.instance
        selector = infer_pipeline_selector(graphql_context, 'no_config_pipeline')
        run_id = make_new_run_id()
        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': {'storage': {'filesystem': {}}},
                    'executionMetadata': {'runId': run_id},
                    'mode': 'default',
                }
            },
        )

        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        assert result.data['launchPipelineExecution']['run']['status'] == 'NOT_STARTED'

        instance.run_launcher.join()

        result = execute_dagster_graphql(
            context=graphql_context, query=RUN_QUERY, variables={'runId': run_id}
        )
        assert result.data['pipelineRunOrError']['__typename'] == 'PipelineRun'
        assert result.data['pipelineRunOrError']['status'] == 'SUCCESS'

        # reexecution
        new_run_id = make_new_run_id()
        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': {'storage': {'filesystem': {}}},
                    'executionMetadata': {
                        'runId': new_run_id,
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                    },
                    'mode': 'default',
                }
            },
        )
        assert result.data['launchPipelineReexecution']['__typename'] == 'LaunchPipelineRunSuccess'

        instance.run_launcher.join()

        result = execute_dagster_graphql(
            context=graphql_context, query=RUN_QUERY, variables={'runId': new_run_id}
        )
        assert result.data['pipelineRunOrError']['__typename'] == 'PipelineRun'
        assert result.data['pipelineRunOrError']['status'] == 'SUCCESS'



from collections import namedtuple

from dagster_graphql.schema.errors import DauphinPipelineConfigValidationInvalid
from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.config.validate import validate_config_from_snap
from dagster.core.errors import DagsterInvalidConfigError
from dagster.core.events import EngineEventData
from dagster.core.execution.api import execute_run
from dagster.core.host_representation.handle import IN_PROCESS_NAME
from dagster.utils.error import SerializableErrorInfo
from dagster.utils.hosted_user_process import recon_pipeline_from_origin

from ..external import get_external_pipeline_or_raise
from ..utils import PipelineSelector, capture_dauphin_error


@capture_dauphin_error
def execute_run_in_graphql_process(
    graphene_info, repository_location_name, repository_name, run_id
):
    '''This indirection is done on purpose to make the logic in the function
    below re-usable. The parent function is wrapped in @capture_dauphin_error, which makes it
    difficult to do exception handling.
    '''
    return _synchronously_execute_run_within_hosted_user_process(
        graphene_info, repository_location_name, repository_name, run_id
    )


RunExecutionInfo = namedtuple('_RunExecutionInfo', 'external_pipeline pipeline_run')


# Welcome, to Greasy Hack Park
#
# Until https://github.com/dagster-io/dagster/issues/2608 is resolved
# the launch_scheduled_execution may be going from user process requesting in to
# a host process, which means it will be incorrectly requesting repository_location_name
# as "<<in_process>>". Below we attempt to workaround this by matching on the other
# identifiers, and replacing repository_location_name if there is a single match.
def _repository_location_name_workaround(
    context, respository_location_name, repository_name, pipeline_name
):
    if respository_location_name != IN_PROCESS_NAME:
        return respository_location_name

    match = None
    for location in context.repository_locations:
        if location.has_repository(repository_name):
            repo = location.get_repository(repository_name)
            if repo.has_pipeline(pipeline_name) and match is None:
                match = location.name
            elif repo.has_pipeline(pipeline_name) and match is not None:
                # multiple matches, return original
                return respository_location_name
    if match:
        return match

    return respository_location_name


def _get_selector_with_workaround(
    context, respository_location_name, repository_name, pipeline_run
):
    return PipelineSelector(
        location_name=_repository_location_name_workaround(
            context, respository_location_name, repository_name, pipeline_run.pipeline_name
        ),
        repository_name=repository_name,
        pipeline_name=pipeline_run.pipeline_name,
        solid_selection=list(pipeline_run.solids_to_execute)
        if pipeline_run.solids_to_execute
        else None,
    )


def get_run_execution_info_for_created_run_or_error(
    graphene_info, repository_location_name, repository_name, run_id
):
    '''
    Previously created run could either be created in a different process *or*
    during the launchScheduledRun call where we want to have a record of
    a run the was created but have invalid configuration
    '''
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.str_param(repository_location_name, 'repository_location_name')
    check.str_param(repository_name, 'repository_name')
    check.str_param(run_id, 'run_id')

    instance = graphene_info.context.instance

    pipeline_run = instance.get_run_by_id(run_id)
    if not pipeline_run:
        return graphene_info.schema.type_named('PipelineRunNotFoundError')(run_id)

    external_pipeline = get_external_pipeline_or_raise(
        graphene_info,
        _get_selector_with_workaround(
            graphene_info.context, repository_location_name, repository_name, pipeline_run
        ),
    )

    validated_config = validate_config_from_snap(
        external_pipeline.config_schema_snapshot,
        external_pipeline.root_config_key_for_mode(pipeline_run.mode),
        pipeline_run.run_config,
    )

    if not validated_config.success:
        # If the config is invalid, we construct a DagsterInvalidConfigError exception and
        # insert it into the event log. We also return a PipelineConfigValidationInvalid user facing
        # graphql error.

        # We currently re-use the engine events machinery to add the error to the event log, but
        # may need to create a new event type and instance method to handle these errors.
        invalid_config_exception = DagsterInvalidConfigError(
            'Error in config for pipeline {}'.format(external_pipeline.name),
            validated_config.errors,
            pipeline_run.run_config,
        )

        instance.report_engine_event(
            str(invalid_config_exception.message),
            pipeline_run,
            EngineEventData.engine_error(
                SerializableErrorInfo(
                    invalid_config_exception.message,
                    [],
                    DagsterInvalidConfigError.__class__.__name__,
                    None,
                )
            ),
        )

        instance.report_run_failed(pipeline_run)

        return DauphinPipelineConfigValidationInvalid.for_validation_errors(
            external_pipeline, validated_config.errors
        )

    return RunExecutionInfo(external_pipeline, pipeline_run)


def _synchronously_execute_run_within_hosted_user_process(
    graphene_info, repository_location_name, repository_name, run_id,
):
    run_info_or_error = get_run_execution_info_for_created_run_or_error(
        graphene_info, repository_location_name, repository_name, run_id
    )

    if not isinstance(run_info_or_error, RunExecutionInfo):
        # if it is not a success the return value is the dauphin error
        return run_info_or_error

    external_pipeline, pipeline_run = run_info_or_error
    recon_pipeline = recon_pipeline_from_origin(external_pipeline.get_origin())
    execute_run(recon_pipeline, pipeline_run, graphene_info.context.instance)
    return graphene_info.schema.type_named('ExecuteRunInProcessSuccess')(
        run=graphene_info.schema.type_named('PipelineRun')(pipeline_run)
    )



from dagster import check
from dagster.core.launcher import RunLauncher
from dagster.serdes import ConfigurableClass


class ExplodingRunLauncher(RunLauncher, ConfigurableClass):
    def __init__(self, inst_data=None):
        self._inst_data = inst_data

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data, config_value):
        return ExplodingRunLauncher(inst_data=inst_data)

    def launch_run(self, instance, run, external_pipeline):
        raise NotImplementedError('The entire purpose of this is to throw on launch')

    def join(self):
        '''Nothing to join on since all executions are synchronous.'''

    def can_terminate(self, run_id):
        return False

    def terminate(self, run_id):
        check.not_implemented('Termination not supported')



from __future__ import absolute_import

from dagster_graphql import dauphin


def create_schema():
    # dauphin depends on import-based side effects
    # pylint: disable=W0611
    from . import (
        assets,
        config_types,
        dagster_types,
        errors,
        execution,
        external,
        paging,
        partition_sets,
        pipelines,
        roots,
        runs,
        schedules,
    )

    return dauphin.create_schema()



from dagster_graphql.test.utils import execute_dagster_graphql

from dagster import repository
from dagster.core.host_representation import external_repository_data_from_def
from dagster.seven import mock

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

# import dagster.api.utils


class TestReloadRepositoriesOutOfProcess(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.readonly_in_memory_instance_out_of_process_env()]
    )
):
    def test_out_of_process_reload_location(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context, RELOAD_REPOSITORY_LOCATION_QUERY, {'repositoryLocationName': 'test'}
        )

        assert result
        assert result.data
        assert result.data['reloadRepositoryLocation']
        assert result.data['reloadRepositoryLocation']['__typename'] == 'RepositoryLocation'
        assert result.data['reloadRepositoryLocation']['name'] == 'test'
        assert result.data['reloadRepositoryLocation']['repositories'] == [{'name': 'test_repo'}]
        assert result.data['reloadRepositoryLocation']['isReloadSupported'] is True

        with mock.patch(
            # note it where the function is *used* that needs to mocked, not
            # where it is defined.
            # see https://docs.python.org/3/library/unittest.mock.html#where-to-patch
            'dagster.api.snapshot_repository.execute_unary_api_cli_command'
        ) as cli_command_mock:

            @repository
            def new_repo():
                return []

            new_repo_data = external_repository_data_from_def(new_repo)

            cli_command_mock.return_value = new_repo_data

            result = execute_dagster_graphql(
                graphql_context,
                RELOAD_REPOSITORY_LOCATION_QUERY,
                {'repositoryLocationName': 'test'},
            )

            assert cli_command_mock.call_count == 1

            assert result.data['reloadRepositoryLocation']['repositories'] == [{'name': 'new_repo'}]


class TestReloadRepositoriesInProcess(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.readonly_in_memory_instance_in_process_env()]
    )
):
    def test_reload_not_supported(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            RELOAD_REPOSITORY_LOCATION_QUERY,
            {'repositoryLocationName': '<<in_process>>'},
        )
        assert result.data['reloadRepositoryLocation']['__typename'] == 'ReloadNotSupported'

    def test_location_not_found(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context, RELOAD_REPOSITORY_LOCATION_QUERY, {'repositoryLocationName': 'nope'},
        )
        assert result.data['reloadRepositoryLocation']['__typename'] == 'RepositoryLocationNotFound'


RELOAD_REPOSITORY_LOCATION_QUERY = '''
mutation ($repositoryLocationName: String!) {
   reloadRepositoryLocation(repositoryLocationName: $repositoryLocationName) {
      __typename
      ... on RepositoryLocation {
        name
        repositories {
            name
        }
        isReloadSupported
      }
   }
}
'''



from collections import OrderedDict, defaultdict

from dagster_graphql.schema.pipelines import DauphinPipeline
from dagster_graphql.schema.solids import (
    DauphinSolidInvocationSite,
    DauphinUsedSolid,
    build_dauphin_solid_handles,
)

from dagster import check
from dagster.core.host_representation import ExternalRepository


def get_solid(repo, name):
    return get_used_solid_map(repo)[name]


def get_solids(repo):
    return get_used_solid_map(repo).values()


def get_used_solid_map(repo):
    check.inst_param(repo, 'repo', ExternalRepository)

    inv_by_def_name = defaultdict(list)
    definitions = []

    for external_pipeline in repo.get_all_external_pipelines():
        for handle in build_dauphin_solid_handles(
            external_pipeline, external_pipeline.dep_structure_index
        ):
            definition = handle.solid.get_dauphin_solid_definition()
            if definition.name not in inv_by_def_name:
                definitions.append(definition)
            inv_by_def_name[definition.name].append(
                DauphinSolidInvocationSite(
                    pipeline=DauphinPipeline(external_pipeline), solidHandle=handle,
                )
            )

    return OrderedDict(
        (
            definition.name,
            DauphinUsedSolid(
                definition=definition,
                invocations=sorted(
                    inv_by_def_name[definition.name],
                    key=lambda i: i.solidHandle.handleID.to_string(),
                ),
            ),
        )
        for definition in sorted(definitions, key=lambda d: d.name)
    )



import csv
from collections import OrderedDict

from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from dagster import (
    AssetMaterialization,
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    PythonObjectDagsterType,
    SolidDefinition,
    dagster_type_loader,
    dagster_type_materializer,
    repository,
)

from .production_query import PRODUCTION_QUERY


@dagster_type_loader(str)
def df_input_schema(_context, path):
    with open(path, 'r') as fd:
        return [OrderedDict(sorted(x.items(), key=lambda x: x[0])) for x in csv.DictReader(fd)]


@dagster_type_materializer(str)
def df_output_schema(_context, path, value):
    with open(path, 'w') as fd:
        writer = csv.DictWriter(fd, fieldnames=value[0].keys())
        writer.writeheader()
        writer.writerows(rowdicts=value)

    return AssetMaterialization.file(path)


PoorMansDataFrame = PythonObjectDagsterType(
    python_type=list,
    name='PoorMansDataFrame',
    loader=df_input_schema,
    materializer=df_output_schema,
)


def test_enum_query(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "pipeline_with_enum_config")

    ENUM_QUERY = '''
    query EnumQuery($selector: PipelineSelector!) {
      runConfigSchemaOrError(selector: $selector) {
        ... on RunConfigSchema {
          allConfigTypes {
            __typename
            key
            ... on EnumConfigType {
              values
              {
                value
                description
              }
            }
          }
        }
      }
    }
    '''

    result = execute_dagster_graphql(graphql_context, ENUM_QUERY, {'selector': selector,},)

    assert not result.errors
    assert result.data

    enum_type_data = None

    for td in result.data['runConfigSchemaOrError']['allConfigTypes']:
        if td['key'] == 'TestEnum':
            enum_type_data = td
            break

    assert enum_type_data
    assert enum_type_data['key'] == 'TestEnum'
    assert enum_type_data['values'] == [
        {'value': 'ENUM_VALUE_ONE', 'description': 'An enum value.'},
        {'value': 'ENUM_VALUE_TWO', 'description': 'An enum value.'},
        {'value': 'ENUM_VALUE_THREE', 'description': 'An enum value.'},
    ]


TYPE_RENDER_QUERY = '''
fragment innerInfo on ConfigType {
  key
  recursiveConfigTypes {
    key
  }
  ... on CompositeConfigType {
    fields {
      name
      configType {
        key
      }
      isRequired
    }
  }
}

query TypeRenderQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      name
      solids {
        name
        definition {
          ... on SolidDefinition {
            configField {
              configType {
                ...innerInfo
                recursiveConfigTypes {
                  ...innerInfo
                }
              }
            }
          }
        }
      }
    }
  }
}
'''


def test_type_rendering(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "more_complicated_nested_config")
    result = execute_dagster_graphql(graphql_context, TYPE_RENDER_QUERY, {'selector': selector})
    assert not result.errors
    assert result.data


def define_circular_dependency_pipeline():
    return PipelineDefinition(
        name='circular_dependency_pipeline',
        solid_defs=[
            SolidDefinition(
                name='csolid',
                input_defs=[InputDefinition('num', PoorMansDataFrame)],
                output_defs=[OutputDefinition(PoorMansDataFrame)],
                compute_fn=lambda *_args: None,
            )
        ],
        dependencies={'csolid': {'num': DependencyDefinition('csolid')}},
    )


@repository
def test_repository():
    return {'pipelines': {'pipeline': define_circular_dependency_pipeline}}


def test_pipeline_or_error_by_name(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "csv_hello_world_two")
    result = execute_dagster_graphql(
        graphql_context,
        '''
    query NamedPipelineQuery($selector: PipelineSelector!) {
        pipelineOrError(params: $selector) {
            ... on Pipeline {
                name
            }
        }
    }''',
        {'selector': selector},
    )

    assert not result.errors
    assert result.data
    assert result.data['pipelineOrError']['name'] == 'csv_hello_world_two'


def test_pipeline_or_error_by_name_not_found(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "foobar")
    result = execute_dagster_graphql(
        graphql_context,
        '''
        query NamedPipelineQuery($selector: PipelineSelector!) {
            pipelineOrError(params: $selector) {
                __typename
                ... on Pipeline {
                    name
                }
            }
        }''',
        {'selector': selector},
    )

    assert not result.errors
    assert result.data
    assert result.data['pipelineOrError']['__typename'] == 'PipelineNotFoundError'


def test_production_query(graphql_context):
    result = execute_dagster_graphql(graphql_context, PRODUCTION_QUERY)

    assert not result.errors
    assert result.data



from collections import defaultdict

from dagster_graphql.cli import execute_query
from dagster_graphql.client.query import EXECUTE_PLAN_MUTATION
from dagster_graphql.client.util import HANDLED_EVENTS, dagster_event_from_dict, parse_raw_log_lines

from dagster import (
    AssetMaterialization,
    Bool,
    DependencyDefinition,
    EventMetadataEntry,
    ExpectationResult,
    InputDefinition,
    Output,
    OutputDefinition,
    PipelineDefinition,
    RetryRequested,
    execute_pipeline,
    lambda_solid,
    solid,
)
from dagster.cli.workspace.cli_target import PythonFileTarget, workspace_from_load_target
from dagster.core.events import STEP_EVENTS, DagsterEventType
from dagster.core.execution.api import create_execution_plan
from dagster.core.host_representation import UserProcessApi
from dagster.core.instance import DagsterInstance


def test_can_handle_all_step_events():
    '''This test is designed to ensure we catch the case when new step events are added, as they
    must be handled by the event parsing, but this does not check that the event parsing works
    correctly.
    '''
    handled = set(HANDLED_EVENTS.values())
    # The distinction between "step events" and "pipeline events" needs to be reexamined
    assert handled == STEP_EVENTS.union(set([DagsterEventType.ENGINE_EVENT]))


def define_test_events_pipeline():
    @solid(output_defs=[OutputDefinition(Bool)])
    def materialization_and_expectation(_context):
        yield AssetMaterialization(
            asset_key='all_types',
            description='a materialization with all metadata types',
            metadata_entries=[
                EventMetadataEntry.text('text is cool', 'text'),
                EventMetadataEntry.url('https://bigty.pe/neato', 'url'),
                EventMetadataEntry.fspath('/tmp/awesome', 'path'),
                EventMetadataEntry.json({'is_dope': True}, 'json'),
            ],
        )
        yield ExpectationResult(success=True, label='row_count', description='passed')
        yield ExpectationResult(True)
        yield Output(True)

    @solid(
        output_defs=[
            OutputDefinition(name='output_one'),
            OutputDefinition(name='output_two', is_required=False),
        ]
    )
    def optional_only_one(_context):  # pylint: disable=unused-argument
        yield Output(output_name='output_one', value=1)

    @solid(input_defs=[InputDefinition('some_input')])
    def should_fail(_context, some_input):  # pylint: disable=unused-argument
        raise Exception('should fail')

    @solid(input_defs=[InputDefinition('some_input')])
    def should_be_skipped(_context, some_input):  # pylint: disable=unused-argument
        pass

    @lambda_solid
    def retries():
        raise RetryRequested()

    return PipelineDefinition(
        name='test_events',
        solid_defs=[
            materialization_and_expectation,
            optional_only_one,
            should_fail,
            should_be_skipped,
            retries,
        ],
        dependencies={
            'optional_only_one': {},
            'should_fail': {
                'some_input': DependencyDefinition(optional_only_one.name, 'output_one')
            },
            'should_be_skipped': {
                'some_input': DependencyDefinition(optional_only_one.name, 'output_two')
            },
        },
    )


def test_pipeline():
    '''just a sanity check to ensure the above pipeline works without layering on graphql'''
    result = execute_pipeline(define_test_events_pipeline(), raise_on_error=False)
    assert result.result_for_solid('materialization_and_expectation').success
    assert not result.result_for_solid('should_fail').success
    assert result.result_for_solid('should_be_skipped').skipped


def test_all_step_events():  # pylint: disable=too-many-locals
    workspace = workspace_from_load_target(
        PythonFileTarget(
            __file__, define_test_events_pipeline.__name__, user_process_api=UserProcessApi.CLI
        )
    )
    pipeline_def = define_test_events_pipeline()
    mode = pipeline_def.get_default_mode_name()
    instance = DagsterInstance.ephemeral()
    execution_plan = create_execution_plan(pipeline_def, mode=mode)
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=pipeline_def, execution_plan=execution_plan, mode=mode
    )
    step_levels = execution_plan.topological_step_levels()

    unhandled_events = STEP_EVENTS.copy()

    # Exclude types that are not step events
    ignored_events = {
        'LogMessageEvent',
        'PipelineStartEvent',
        'PipelineSuccessEvent',
        'PipelineInitFailureEvent',
        'PipelineFailureEvent',
    }

    event_counts = defaultdict(int)

    for step_level in step_levels:
        for step in step_level:

            variables = {
                'executionParams': {
                    'selector': {
                        'repositoryLocationName': 'test_events',
                        'repositoryName': '<<unnamed>>',
                        'pipelineName': pipeline_def.name,
                    },
                    'runConfigData': {'storage': {'filesystem': {}}},
                    'mode': mode,
                    'executionMetadata': {'runId': pipeline_run.run_id},
                    'stepKeys': [step.key],
                },
            }
            res = execute_query(workspace, EXECUTE_PLAN_MUTATION, variables, instance=instance,)

            # go through the same dict, decrement all the event records we've seen from the GraphQL
            # response
            if not res.get('errors'):
                assert 'data' in res, res
                assert 'executePlan' in res['data'], res
                assert 'stepEvents' in res['data']['executePlan'], res
                step_events = res['data']['executePlan']['stepEvents']

                events = [
                    dagster_event_from_dict(e, pipeline_def.name)
                    for e in step_events
                    if e['__typename'] not in ignored_events
                ]

                for event in events:
                    if event.step_key:
                        key = event.step_key + '.' + event.event_type_value
                    else:
                        key = event.event_type_value
                    event_counts[key] -= 1
                unhandled_events -= {DagsterEventType(e.event_type_value) for e in events}
            else:
                raise Exception(res['errors'])

    # build up a dict, incrementing all the event records we've produced in the run storage
    logs = instance.all_logs(pipeline_run.run_id)
    for log in logs:
        if not log.dagster_event or (
            DagsterEventType(log.dagster_event.event_type_value)
            not in STEP_EVENTS.union(set([DagsterEventType.ENGINE_EVENT]))
        ):
            continue
        if log.dagster_event.step_key:
            key = log.dagster_event.step_key + '.' + log.dagster_event.event_type_value
        else:
            key = log.dagster_event.event_type_value
        event_counts[key] += 1

    # Ensure we've processed all the events that were generated in the run storage
    assert sum(event_counts.values()) == 0

    # Ensure we've handled the universe of event types
    # Why are these retry events not handled? Because right now there is no way to configure retries
    # on executePlan -- this needs to change, and we should separate the ExecutionParams that get
    # sent to executePlan fromm those that get sent to startPipelineExecution and friends
    assert unhandled_events == {DagsterEventType.STEP_UP_FOR_RETRY, DagsterEventType.STEP_RESTARTED}


def test_parse_raw_log_lines():
    raw_log_lines = [
        'Some extraneous log line',
        'Another extraneous log line, JSON data is on the next line',
        '{"data": {"executePlan": {"__typename": "ExecutePlanSuccess"}}}',
    ]

    result = {'data': {'executePlan': {'__typename': 'ExecutePlanSuccess'}}}
    assert parse_raw_log_lines(raw_log_lines) == result



from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

RUN_QUERY = '''
query RunQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on PipelineRun {
      status
      stats {
        ... on PipelineRunStatsSnapshot {
          stepsSucceeded
        }
      }
    }
  }
}
'''


class TestBasicLaunch(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.sqlite_with_cli_api_run_launcher_in_process_env()]
    )
):
    def test_run_launcher(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'no_config_pipeline')
        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': selector, 'mode': 'default'}},
        )

        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        assert result.data['launchPipelineExecution']['run']['status'] == 'NOT_STARTED'

        run_id = result.data['launchPipelineExecution']['run']['runId']

        graphql_context.instance.run_launcher.join()

        result = execute_dagster_graphql(
            context=graphql_context, query=RUN_QUERY, variables={'runId': run_id}
        )
        assert result.data['pipelineRunOrError']['__typename'] == 'PipelineRun'
        assert result.data['pipelineRunOrError']['status'] == 'SUCCESS'

    def test_run_launcher_subset(self, graphql_context):
        selector = infer_pipeline_selector(
            graphql_context, 'more_complicated_config', ['noop_solid']
        )
        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': selector, 'mode': 'default',}},
        )

        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        assert result.data['launchPipelineExecution']['run']['status'] == 'NOT_STARTED'

        run_id = result.data['launchPipelineExecution']['run']['runId']

        graphql_context.instance.run_launcher.join()

        result = execute_dagster_graphql(
            context=graphql_context, query=RUN_QUERY, variables={'runId': run_id}
        )
        assert result.data['pipelineRunOrError']['__typename'] == 'PipelineRun'
        assert result.data['pipelineRunOrError']['status'] == 'SUCCESS'
        assert result.data['pipelineRunOrError']['stats']['stepsSucceeded'] == 1



import json

from dagster_graphql.test.utils import infer_pipeline_selector

from dagster.utils import file_relative_path

from .graphql_context_test_suite import ExecutingGraphQLContextTestMatrix
from .utils import sync_execute_get_events


def get_expectation_results(logs, solid_name):
    def _f():
        for log in logs:
            if log['__typename'] == 'StepExpectationResultEvent' and log[
                'stepKey'
            ] == '{}.compute'.format(solid_name):
                yield log

    return list(_f())


def get_expectation_result(logs, solid_name):
    expt_results = get_expectation_results(logs, solid_name)
    if len(expt_results) != 1:
        raise Exception('Only expected one expectation result')
    return expt_results[0]


class TestExpectations(ExecutingGraphQLContextTestMatrix):
    def test_basic_expectations_within_compute_step_events(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, 'pipeline_with_expectations')
        logs = sync_execute_get_events(
            context=graphql_context,
            variables={'executionParams': {'selector': selector, 'mode': 'default'}},
        )

        emit_failed_expectation_event = get_expectation_result(logs, 'emit_failed_expectation')
        assert emit_failed_expectation_event['expectationResult']['success'] is False
        assert emit_failed_expectation_event['expectationResult']['description'] == 'Failure'
        failed_result_metadata = json.loads(
            emit_failed_expectation_event['expectationResult']['metadataEntries'][0]['jsonString']
        )
        assert emit_failed_expectation_event['expectationResult']['label'] == 'always_false'

        assert failed_result_metadata == {'reason': 'Relentless pessimism.'}

        emit_successful_expectation_event = get_expectation_result(
            logs, 'emit_successful_expectation'
        )

        assert emit_successful_expectation_event['expectationResult']['success'] is True
        assert emit_successful_expectation_event['expectationResult']['description'] == 'Successful'
        assert emit_successful_expectation_event['expectationResult']['label'] == 'always_true'
        successful_result_metadata = json.loads(
            emit_successful_expectation_event['expectationResult']['metadataEntries'][0][
                'jsonString'
            ]
        )

        assert successful_result_metadata == {'reason': 'Just because.'}

        emit_no_metadata = get_expectation_result(logs, 'emit_successful_expectation_no_metadata')
        assert not emit_no_metadata['expectationResult']['metadataEntries']

        snapshot.assert_match(get_expectation_results(logs, 'emit_failed_expectation'))
        snapshot.assert_match(get_expectation_results(logs, 'emit_successful_expectation'))
        snapshot.assert_match(
            get_expectation_results(logs, 'emit_successful_expectation_no_metadata')
        )

    def test_basic_input_output_expectations(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world_with_expectations')
        logs = sync_execute_get_events(
            context=graphql_context,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': {
                        'solids': {
                            'sum_solid': {
                                'inputs': {'num': file_relative_path(__file__, '../data/num.csv')}
                            }
                        }
                    },
                    'mode': 'default',
                }
            },
        )

        expectation_results = get_expectation_results(logs, 'df_expectations_solid')
        assert len(expectation_results) == 2

        snapshot.assert_match(expectation_results)



from __future__ import absolute_import

import sys
import time

from dagster_graphql.implementation.fetch_runs import is_config_valid
from dagster_graphql.schema.pipelines import DauphinPipeline
from dagster_graphql.schema.runs import (
    from_compute_log_file,
    from_dagster_event_record,
    from_event_record,
)
from graphql.execution.base import ResolveInfo
from rx import Observable

from dagster import check
from dagster.core.definitions import ExecutablePipeline
from dagster.core.definitions.schedule import ScheduleExecutionContext
from dagster.core.errors import (
    DagsterInvalidConfigError,
    ScheduleExecutionError,
    user_code_error_boundary,
)
from dagster.core.events import DagsterEventType
from dagster.core.execution.retries import Retries
from dagster.core.host_representation import ExternalExecutionPlan
from dagster.core.scheduler import ScheduleTickStatus
from dagster.core.scheduler.scheduler import ScheduleTickData
from dagster.core.storage.compute_log_manager import ComputeIOType
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus
from dagster.core.system_config.objects import ExecutionConfig
from dagster.serdes import serialize_dagster_namedtuple
from dagster.utils.error import serializable_error_info_from_exc_info

from ..external import (
    ExternalPipeline,
    ensure_valid_config,
    ensure_valid_step_keys,
    get_external_pipeline_or_raise,
)
from ..pipeline_run_storage import PipelineRunObservableSubscribe
from ..resume_retry import get_retry_steps_from_execution_plan
from ..utils import ExecutionParams, UserFacingGraphQLError, capture_dauphin_error
from .launch_execution import launch_pipeline_execution, launch_pipeline_reexecution


@capture_dauphin_error
def terminate_pipeline_execution(graphene_info, run_id):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.str_param(run_id, 'run_id')

    instance = graphene_info.context.instance
    run = instance.get_run_by_id(run_id)

    if not run:
        return graphene_info.schema.type_named('PipelineRunNotFoundError')(run_id)

    dauphin_run = graphene_info.schema.type_named('PipelineRun')(run)

    if run.status != PipelineRunStatus.STARTED:
        return graphene_info.schema.type_named('TerminatePipelineExecutionFailure')(
            run=dauphin_run,
            message='Run {run_id} is not in a started state. Current status is {status}'.format(
                run_id=run.run_id, status=run.status.value
            ),
        )

    can_not_term = graphene_info.schema.type_named('TerminatePipelineExecutionFailure')(
        run=dauphin_run, message='Unable to terminate run {run_id}'.format(run_id=run.run_id)
    )

    if (
        graphene_info.context.instance.run_launcher
        and graphene_info.context.instance.run_launcher.can_terminate(run_id)
    ):
        if not graphene_info.context.instance.run_launcher.terminate(run_id):
            return can_not_term

        return graphene_info.schema.type_named('TerminatePipelineExecutionSuccess')(dauphin_run)

    else:
        return can_not_term


@capture_dauphin_error
def delete_pipeline_run(graphene_info, run_id):
    instance = graphene_info.context.instance

    if not instance.has_run(run_id):
        return graphene_info.schema.type_named('PipelineRunNotFoundError')(run_id)

    instance.delete_run(run_id)

    return graphene_info.schema.type_named('DeletePipelineRunSuccess')(run_id)


def get_pipeline_run_observable(graphene_info, run_id, after=None):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.str_param(run_id, 'run_id')
    check.opt_int_param(after, 'after')
    instance = graphene_info.context.instance
    run = instance.get_run_by_id(run_id)

    if not run:

        def _get_error_observable(observer):
            observer.on_next(
                graphene_info.schema.type_named('PipelineRunLogsSubscriptionFailure')(
                    missingRunId=run_id, message='Could not load run with id {}'.format(run_id)
                )
            )

        return Observable.create(_get_error_observable)  # pylint: disable=E1101

    # pylint: disable=E1101
    return Observable.create(
        PipelineRunObservableSubscribe(instance, run_id, after_cursor=after)
    ).map(
        lambda events: graphene_info.schema.type_named('PipelineRunLogsSubscriptionSuccess')(
            run=graphene_info.schema.type_named('PipelineRun')(run),
            messages=[from_event_record(event, run.pipeline_name) for event in events],
        )
    )


def get_compute_log_observable(graphene_info, run_id, step_key, io_type, cursor=None):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.str_param(run_id, 'run_id')
    check.str_param(step_key, 'step_key')
    check.inst_param(io_type, 'io_type', ComputeIOType)
    check.opt_str_param(cursor, 'cursor')

    return graphene_info.context.instance.compute_log_manager.observable(
        run_id, step_key, io_type, cursor
    ).map(lambda update: from_compute_log_file(graphene_info, update))


@capture_dauphin_error
def do_execute_plan(graphene_info, execution_params, retries):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(execution_params, 'execution_params', ExecutionParams)
    check.opt_inst_param(retries, 'retries', Retries)

    external_pipeline = get_external_pipeline_or_raise(graphene_info, execution_params.selector)
    ensure_valid_config(
        external_pipeline=external_pipeline,
        mode=execution_params.mode,
        run_config=execution_params.run_config,
    )
    return _do_execute_plan(graphene_info, execution_params, external_pipeline, retries)


def _do_execute_plan(graphene_info, execution_params, external_pipeline, retries):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(execution_params, 'execution_params', ExecutionParams)
    check.inst_param(external_pipeline, 'external_pipeline', ExternalPipeline)
    check.opt_inst_param(retries, 'retries', Retries)

    run_id = execution_params.execution_metadata.run_id

    mode = execution_params.mode or external_pipeline.get_default_mode_name()

    pipeline_run = graphene_info.context.instance.get_run_by_id(run_id)

    external_execution_plan = graphene_info.context.get_external_execution_plan(
        external_pipeline=external_pipeline,
        run_config=execution_params.run_config,
        mode=mode,
        step_keys_to_execute=None,
    )

    if not pipeline_run:
        # TODO switch to raising a UserFacingError if the run_id cannot be found
        # https://github.com/dagster-io/dagster/issues/1876
        pipeline_run = graphene_info.context.instance.create_run(
            pipeline_name=external_pipeline.name,
            run_id=run_id,
            run_config=execution_params.run_config,
            mode=mode,
            solids_to_execute=None,
            step_keys_to_execute=None,
            status=None,
            tags=execution_params.execution_metadata.tags or {},
            root_run_id=None,
            parent_run_id=None,
            pipeline_snapshot=external_pipeline.pipeline_snapshot,
            execution_plan_snapshot=external_execution_plan.execution_plan_snapshot,
            parent_pipeline_snapshot=external_pipeline.parent_pipeline_snapshot,
        )

    ensure_valid_step_keys(external_execution_plan, execution_params.step_keys)

    if execution_params.step_keys:
        external_execution_plan = graphene_info.context.get_external_execution_plan(
            external_pipeline=external_pipeline,
            run_config=execution_params.run_config,
            mode=mode,
            step_keys_to_execute=execution_params.step_keys,
        )

    event_logs = []

    def _on_event_record(record):
        if record.is_dagster_event:
            event_logs.append(record)

    graphene_info.context.instance.add_event_listener(run_id, _on_event_record)

    graphene_info.context.execute_plan(
        external_pipeline=external_pipeline,
        run_config=execution_params.run_config,
        pipeline_run=pipeline_run,
        step_keys_to_execute=execution_params.step_keys,
        retries=retries,
    )

    def to_graphql_event(event_record):
        return from_dagster_event_record(event_record, external_pipeline.name)

    return graphene_info.schema.type_named('ExecutePlanSuccess')(
        pipeline=DauphinPipeline(external_pipeline),
        has_failures=any(
            er
            for er in event_logs
            if er.is_dagster_event and er.dagster_event.event_type == DagsterEventType.STEP_FAILURE
        ),
        step_events=list(map(to_graphql_event, event_logs)),
        raw_event_records=list(map(serialize_dagster_namedtuple, event_logs)),
    )



from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

GET_ASSET_KEY_QUERY = '''
{
    assetsOrError {
        __typename
        ...on AssetConnection {
            nodes {
                key {
                    path
                }
            }
        }
    }
}
'''

GET_ASSET_MATERIALIZATION = '''
    query AssetQuery($assetKey: AssetKeyInput!) {
        assetOrError(assetKey: $assetKey) {
            ... on Asset {
                assetMaterializations(limit: 1) {
                    materializationEvent {
                        materialization {
                            label
                        }
                    }
                }
            }
        }
    }
'''

GET_ASSET_RUNS = '''
    query AssetRunsQuery($assetKey: AssetKeyInput!) {
        assetOrError(assetKey: $assetKey) {
            ... on Asset {
                runs {
                    runId
                }
            }
        }
    }
'''


class TestAssetAwareEventLog(
    make_graphql_context_test_suite(
        context_variants=[
            GraphQLContextVariant.in_memory_instance_in_process_env(),
            GraphQLContextVariant.asset_aware_sqlite_instance_in_process_env(),
            GraphQLContextVariant.postgres_with_sync_run_launcher_in_process_env(),
        ]
    )
):
    def test_get_all_asset_keys(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, 'multi_asset_pipeline')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': selector, 'mode': 'default'}},
        )
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'

        result = execute_dagster_graphql(graphql_context, GET_ASSET_KEY_QUERY)
        assert result.data
        assert result.data['assetsOrError']
        assert result.data['assetsOrError']['nodes']

        # sort by materialization asset key to keep list order is consistent for snapshot
        result.data['assetsOrError']['nodes'].sort(key=lambda e: e['key']['path'][0])

        snapshot.assert_match(result.data)

    def test_get_asset_key_materialization(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, 'single_asset_pipeline')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': selector, 'mode': 'default'}},
        )
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        result = execute_dagster_graphql(
            graphql_context, GET_ASSET_MATERIALIZATION, variables={'assetKey': {'path': ['a']}}
        )
        assert result.data
        snapshot.assert_match(result.data)

    def test_get_asset_runs(self, graphql_context):
        single_selector = infer_pipeline_selector(graphql_context, 'single_asset_pipeline')
        multi_selector = infer_pipeline_selector(graphql_context, 'multi_asset_pipeline')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': single_selector, 'mode': 'default'}},
        )
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        single_run_id = result.data['launchPipelineExecution']['run']['runId']

        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': multi_selector, 'mode': 'default'}},
        )
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        multi_run_id = result.data['launchPipelineExecution']['run']['runId']

        result = execute_dagster_graphql(
            graphql_context, GET_ASSET_RUNS, variables={'assetKey': {'path': ['a']}}
        )
        assert result.data
        fetched_runs = [run['runId'] for run in result.data['assetOrError']['runs']]
        assert len(fetched_runs) == 2
        assert multi_run_id in fetched_runs
        assert single_run_id in fetched_runs



from future.standard_library import install_aliases  # isort:skip

install_aliases()  # isort:skip

import signal

import click
import requests
from graphql import graphql
from graphql.execution.executors.gevent import GeventExecutor
from graphql.execution.executors.sync import SyncExecutor

from dagster import check, seven
from dagster.cli.workspace import workspace_target_argument
from dagster.cli.workspace.cli_target import get_workspace_from_kwargs
from dagster.cli.workspace.workspace import Workspace
from dagster.core.host_representation.repository_location import RepositoryLocation
from dagster.core.instance import DagsterInstance
from dagster.seven import urljoin, urlparse
from dagster.utils import DEFAULT_REPOSITORY_YAML_FILENAME
from dagster.utils.log import get_stack_trace_array

from .client.query import (
    EXECUTE_PLAN_MUTATION,
    EXECUTE_RUN_IN_PROCESS_MUTATION,
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
)
from .implementation.context import DagsterGraphQLContext
from .schema import create_schema
from .version import __version__

# TODO we may want to start extracting shared copy like this to some central location.
REPO_TARGET_WARNING = (
    'Can only use ONE of --repository-yaml/-y, --python-file/-f, --module-name/-m.'
)


def create_dagster_graphql_cli():
    return ui


def execute_query(workspace, query, variables=None, use_sync_executor=False, instance=None):
    check.inst_param(workspace, 'workspace', Workspace)
    check.str_param(query, 'query')
    check.opt_dict_param(variables, 'variables')
    instance = (
        check.inst_param(instance, 'instance', DagsterInstance)
        if instance
        else DagsterInstance.get()
    )
    check.bool_param(use_sync_executor, 'use_sync_executor')

    query = query.strip('\'" \n\t')

    locations = [RepositoryLocation.from_handle(x) for x in workspace.repository_location_handles]

    context = DagsterGraphQLContext(locations=locations, instance=instance, version=__version__,)

    executor = SyncExecutor() if use_sync_executor else GeventExecutor()

    result = graphql(
        request_string=query,
        schema=create_schema(),
        context_value=context,
        variable_values=variables,
        executor=executor,
    )

    result_dict = result.to_dict()

    context.drain_outstanding_executions()

    # Here we detect if this is in fact an error response
    # If so, we iterate over the result_dict and the original result
    # which contains a GraphQLError. If that GraphQL error contains
    # an original_error property (which is the exception the resolver
    # has thrown, typically) we serialize the stack trace of that exception
    # in the 'stack_trace' property of each error to ease debugging

    if 'errors' in result_dict:
        check.invariant(len(result_dict['errors']) == len(result.errors))
        for python_error, error_dict in zip(result.errors, result_dict['errors']):
            if hasattr(python_error, 'original_error') and python_error.original_error:
                error_dict['stack_trace'] = get_stack_trace_array(python_error.original_error)

    return result_dict


def execute_query_from_cli(workspace, query, variables=None, output=None):
    check.inst_param(workspace, 'workspace', Workspace)
    check.str_param(query, 'query')
    check.opt_str_param(variables, 'variables')
    check.opt_str_param(output, 'output')

    query = query.strip('\'" \n\t')

    result_dict = execute_query(
        workspace, query, variables=seven.json.loads(variables) if variables else None
    )
    str_res = seven.json.dumps(result_dict)

    # Since this the entry point for CLI execution, some tests depend on us putting the result on
    # stdout
    if output:
        check.str_param(output, 'output')
        with open(output, 'w') as f:
            f.write(str_res + '\n')
    else:
        print(str_res)  # pylint: disable=print-call

    return str_res


def execute_query_against_remote(host, query, variables):
    parsed_url = urlparse(host)
    if not (parsed_url.scheme and parsed_url.netloc):
        raise click.UsageError(
            'Host {host} is not a valid URL. Host URL should include scheme ie http://localhost'.format(
                host=host
            )
        )

    sanity_check = requests.get(urljoin(host, '/dagit_info'))
    sanity_check.raise_for_status()
    if 'dagit' not in sanity_check.text:
        raise click.UsageError(
            'Host {host} failed sanity check. It is not a dagit server.'.format(host=host)
        )

    response = requests.post(
        urljoin(host, '/graphql'), params={'query': query, 'variables': variables}
    )
    response.raise_for_status()
    str_res = response.json()
    return str_res


PREDEFINED_QUERIES = {
    'executeRunInProcess': EXECUTE_RUN_IN_PROCESS_MUTATION,
    'executePlan': EXECUTE_PLAN_MUTATION,
    'launchPipelineExecution': LAUNCH_PIPELINE_EXECUTION_MUTATION,
}


@workspace_target_argument
@click.command(
    name='ui',
    help=(
        'Run a GraphQL query against the dagster interface to a specified repository or pipeline.'
        '\n\n{warning}'.format(warning=REPO_TARGET_WARNING)
    )
    + (
        '\n\n Examples:'
        '\n\n1. dagster-graphql'
        '\n\n2. dagster-graphql -y path/to/{default_filename}'
        '\n\n3. dagster-graphql -f path/to/file.py -n define_repo'
        '\n\n4. dagster-graphql -m some_module -n define_repo'
        '\n\n5. dagster-graphql -f path/to/file.py -n define_pipeline'
        '\n\n6. dagster-graphql -m some_module -n define_pipeline'
    ).format(default_filename=DEFAULT_REPOSITORY_YAML_FILENAME),
)
@click.version_option(version=__version__)
@click.option(
    '--text', '-t', type=click.STRING, help='GraphQL document to execute passed as a string'
)
@click.option(
    '--file', '-f', type=click.File(), help='GraphQL document to execute passed as a file'
)
@click.option(
    '--predefined',
    '-p',
    type=click.Choice(PREDEFINED_QUERIES.keys()),
    help='GraphQL document to execute, from a predefined set provided by dagster-graphql.',
)
@click.option(
    '--variables',
    '-v',
    type=click.STRING,
    help='A JSON encoded string containing the variables for GraphQL execution.',
)
@click.option(
    '--remote',
    '-r',
    type=click.STRING,
    help='A URL for a remote instance running dagit server to send the GraphQL request to.',
)
@click.option(
    '--output',
    '-o',
    type=click.STRING,
    help='A file path to store the GraphQL response to. This flag is useful when making pipeline '
    'execution queries, since pipeline execution causes logs to print to stdout and stderr.',
)
@click.option(
    '--remap-sigterm', is_flag=True, default=False, help='Remap SIGTERM signal to SIGINT handler',
)
def ui(text, file, predefined, variables, remote, output, remap_sigterm, **kwargs):
    query = None
    if text is not None and file is None and predefined is None:
        query = text.strip('\'" \n\t')
    elif file is not None and text is None and predefined is None:
        query = file.read()
    elif predefined is not None and text is None and file is None:
        query = PREDEFINED_QUERIES[predefined]
    else:
        raise click.UsageError(
            'Must select one and only one of text (-t), file (-f), or predefined (-p) '
            'to select GraphQL document to execute.'
        )

    if remap_sigterm:
        signal.signal(signal.SIGTERM, signal.getsignal(signal.SIGINT))

    if remote:
        res = execute_query_against_remote(remote, query, variables)
        print(res)  # pylint: disable=print-call
    else:
        workspace = get_workspace_from_kwargs(kwargs)
        execute_query_from_cli(workspace, query, variables, output)


cli = create_dagster_graphql_cli()


def main():
    # click magic
    cli(obj={})  # pylint:disable=E1120



import os
import time

from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from dagster import execute_pipeline
from dagster.utils import safe_tempfile_path

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite
from .setup import main_repo_location_name

RUN_CANCELLATION_QUERY = '''
mutation($runId: String!) {
  terminatePipelineExecution(runId: $runId){
    __typename
    ... on TerminatePipelineExecutionSuccess{
      run {
        runId
      }
    }
    ... on TerminatePipelineExecutionFailure {
      run {
        runId
      }
      message
    }
    ... on PipelineRunNotFoundError {
      runId
    }
  }
}
'''


class TestRunVariantTermination(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.sqlite_with_cli_api_run_launcher_in_process_env()]
    )
):
    def test_basic_termination(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'infinite_loop_pipeline')
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    'executionParams': {
                        'selector': selector,
                        'mode': 'default',
                        'runConfigData': {'solids': {'loop': {'config': {'file': path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert (
                result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
            )
            run_id = result.data['launchPipelineExecution']['run']['runId']

            assert run_id

            # ensure the execution has happened
            while not os.path.exists(path):
                time.sleep(0.1)

            result = execute_dagster_graphql(
                graphql_context, RUN_CANCELLATION_QUERY, variables={'runId': run_id}
            )

            assert (
                result.data['terminatePipelineExecution']['__typename']
                == 'TerminatePipelineExecutionSuccess'
            )

    def test_run_not_found(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context, RUN_CANCELLATION_QUERY, variables={'runId': 'nope'}
        )
        assert result.data['terminatePipelineExecution']['__typename'] == 'PipelineRunNotFoundError'

    def test_terminate_failed(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'infinite_loop_pipeline')
        with safe_tempfile_path() as path:
            old_terminate = graphql_context.instance.run_launcher.terminate
            graphql_context.instance.run_launcher.terminate = lambda _run_id: False
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    'executionParams': {
                        'selector': selector,
                        'mode': 'default',
                        'runConfigData': {'solids': {'loop': {'config': {'file': path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert (
                result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
            )
            run_id = result.data['launchPipelineExecution']['run']['runId']
            # ensure the execution has happened
            while not os.path.exists(path):
                time.sleep(0.1)

            result = execute_dagster_graphql(
                graphql_context, RUN_CANCELLATION_QUERY, variables={'runId': run_id}
            )
            assert (
                result.data['terminatePipelineExecution']['__typename']
                == 'TerminatePipelineExecutionFailure'
            )
            assert result.data['terminatePipelineExecution']['message'].startswith(
                'Unable to terminate run'
            )

            graphql_context.instance.run_launcher.terminate = old_terminate

            result = execute_dagster_graphql(
                graphql_context, RUN_CANCELLATION_QUERY, variables={'runId': run_id}
            )

            assert (
                result.data['terminatePipelineExecution']['__typename']
                == 'TerminatePipelineExecutionSuccess'
            )

    def test_run_finished(self, graphql_context):
        instance = graphql_context.instance
        pipeline = graphql_context.get_repository_location(
            main_repo_location_name()
        ).get_reconstructable_pipeline('noop_pipeline')

        pipeline_result = execute_pipeline(pipeline, instance=instance)
        assert pipeline_result.success
        assert pipeline_result.run_id

        time.sleep(0.05)  # guarantee execution finish

        result = execute_dagster_graphql(
            graphql_context, RUN_CANCELLATION_QUERY, variables={'runId': pipeline_result.run_id}
        )

        assert (
            result.data['terminatePipelineExecution']['__typename']
            == 'TerminatePipelineExecutionFailure'
        )
        assert (
            'is not in a started state. Current status is SUCCESS'
            in result.data['terminatePipelineExecution']['message']
        )



from collections import OrderedDict

from dagster_graphql.test.utils import execute_dagster_graphql, infer_repository_selector

from .graphql_context_test_suite import ReadonlyGraphQLContextTestMatrix

GET_PARTITION_SETS_FOR_PIPELINE_QUERY = '''
    query PartitionSetsQuery($repositorySelector: RepositorySelector!, $pipelineName: String!) {
        partitionSetsOrError(repositorySelector: $repositorySelector, pipelineName: $pipelineName) {
            __typename
            ...on PartitionSets {
                results {
                    name
                    pipelineName
                    solidSelection
                    mode
                }
            }
            ... on PythonError {
                message
                stack
            }
            ...on PipelineNotFoundError {
                message
            }
        }
    }
'''

GET_PARTITION_SET_QUERY = '''
    query PartitionSetQuery($repositorySelector: RepositorySelector!, $partitionSetName: String!) {
        partitionSetOrError(repositorySelector: $repositorySelector, partitionSetName: $partitionSetName) {
            __typename
            ... on PythonError {
                message
                stack
            }
            ...on PartitionSet {
                name
                pipelineName
                solidSelection
                mode
                partitionsOrError {
                    ... on Partitions {
                        results {
                            name
                            tagsOrError {
                                __typename
                            }
                            runConfigOrError {
                                ... on PartitionRunConfig {
                                    yaml
                                }
                            }
                        }
                    }
                }
            }
        }
    }
'''

GET_PARTITION_SET_TAGS_QUERY = '''
    query PartitionSetQuery($repositorySelector: RepositorySelector!, $partitionSetName: String!) {
        partitionSetOrError(repositorySelector: $repositorySelector, partitionSetName: $partitionSetName) {
            ...on PartitionSet {
                partitionsOrError(limit: 1) {
                    ... on Partitions {
                        results {
                            tagsOrError {
                                ... on PartitionTags {
                                    results {
                                        key
                                        value
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
'''


class TestPartitionSets(ReadonlyGraphQLContextTestMatrix):
    def test_get_partition_sets_for_pipeline(self, graphql_context, snapshot):
        selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            GET_PARTITION_SETS_FOR_PIPELINE_QUERY,
            variables={'repositorySelector': selector, 'pipelineName': 'no_config_pipeline'},
        )

        assert result.data
        snapshot.assert_match(result.data)

        invalid_pipeline_result = execute_dagster_graphql(
            graphql_context,
            GET_PARTITION_SETS_FOR_PIPELINE_QUERY,
            variables={'repositorySelector': selector, 'pipelineName': 'invalid_pipeline'},
        )

        assert invalid_pipeline_result.data
        snapshot.assert_match(invalid_pipeline_result.data)

    def test_get_partition_set(self, graphql_context, snapshot):
        selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            GET_PARTITION_SET_QUERY,
            variables={'partitionSetName': 'integer_partition', 'repositorySelector': selector},
        )

        assert result.data
        snapshot.assert_match(result.data)

        invalid_partition_set_result = execute_dagster_graphql(
            graphql_context,
            GET_PARTITION_SET_QUERY,
            variables={'partitionSetName': 'invalid_partition', 'repositorySelector': selector},
        )

        assert (
            invalid_partition_set_result.data['partitionSetOrError']['__typename']
            == 'PartitionSetNotFoundError'
        )
        assert invalid_partition_set_result.data

        snapshot.assert_match(invalid_partition_set_result.data)

    def test_get_partition_tags(self, graphql_context):
        selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            GET_PARTITION_SET_TAGS_QUERY,
            variables={'partitionSetName': 'integer_partition', 'repositorySelector': selector},
        )

        assert not result.errors
        assert result.data
        partitions = result.data['partitionSetOrError']['partitionsOrError']['results']
        assert len(partitions) == 1
        sorted_items = sorted(partitions[0]['tagsOrError']['results'], key=lambda item: item['key'])
        tags = OrderedDict({item['key']: item['value'] for item in sorted_items})
        assert tags == {
            'foo': '0',
            'dagster/partition': '0',
            'dagster/partition_set': 'integer_partition',
        }



import json
import os
import time
from contextlib import contextmanager

from click.testing import CliRunner
from dagster_graphql.cli import ui

from dagster import (
    InputDefinition,
    Int,
    OutputDefinition,
    ScheduleDefinition,
    lambda_solid,
    pipeline,
    repository,
    seven,
    solid,
)
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.utils import file_relative_path


@contextmanager
def dagster_cli_runner():
    with seven.TemporaryDirectory() as dagster_home_temp:
        yield CliRunner(env={'DAGSTER_HOME': dagster_home_temp})


@lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
def add_one(num):
    return num + 1


@lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
def mult_two(num):
    return num * 2


@pipeline
def math():
    mult_two(add_one())


@solid(config_schema={'gimme': str})
def needs_config(context):
    return context.solid_config['gimme']


@lambda_solid
def no_config():
    return 'ok'


@pipeline
def subset_test():
    no_config()
    needs_config()


def define_schedules():
    math_hourly_schedule = ScheduleDefinition(
        name="math_hourly_schedule",
        cron_schedule="0 0 * * *",
        pipeline_name="math",
        run_config={'solids': {'add_one': {'inputs': {'num': {'value': 123}}}}},
    )

    return [math_hourly_schedule]


@repository
def test():
    return [math, subset_test] + define_schedules()


def test_basic_introspection():
    query = '{ __schema { types { name } } }'

    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ['-w', workspace_path, '-t', query])
        assert result.exit_code == 0

        result_data = json.loads(result.output)
        assert result_data['data']


def test_basic_repositories():
    query = '{ repositoriesOrError { ... on RepositoryConnection { nodes { name } } } }'

    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ['-w', workspace_path, '-t', query])

        assert result.exit_code == 0

        result_data = json.loads(result.output)
        assert result_data['data']['repositoriesOrError']['nodes']


def test_basic_variables():
    query = '''
    query FooBar($pipelineName: String! $repositoryName: String! $repositoryLocationName: String!){
        pipelineOrError(params:{pipelineName: $pipelineName repositoryName: $repositoryName repositoryLocationName: $repositoryLocationName})
        { ... on Pipeline { name } }
    }
    '''
    variables = '{"pipelineName": "math", "repositoryName": "test", "repositoryLocationName": "<<in_process>>"}'
    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ['-w', workspace_path, '-v', variables, '-t', query])

        assert result.exit_code == 0

        result_data = json.loads(result.output)
        assert result_data['data']['pipelineOrError']['name'] == 'math'


LAUNCH_PIPELINE_EXECUTION_QUERY = '''
mutation ($executionParams: ExecutionParams!) {
    launchPipelineExecution(executionParams: $executionParams) {
        __typename
        ... on LaunchPipelineRunSuccess {
            run {
                runId
                pipeline { ...on PipelineReference { name } }
            }
        }
        ... on PipelineConfigValidationInvalid {
            pipelineName
            errors { message }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
        ... on PythonError {
            message
            stack
        }
    }
}
'''


def test_start_execution_text():
    variables = seven.json.dumps(
        {
            'executionParams': {
                'selector': {
                    'repositoryLocationName': '<<in_process>>',
                    'repositoryName': 'test',
                    'pipelineName': 'math',
                },
                'runConfigData': {'solids': {'add_one': {'inputs': {'num': {'value': 123}}}}},
                'mode': 'default',
            }
        }
    )

    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')

    with dagster_cli_runner() as runner:
        result = runner.invoke(
            ui, ['-w', workspace_path, '-v', variables, '-t', LAUNCH_PIPELINE_EXECUTION_QUERY]
        )

        assert result.exit_code == 0

        try:
            result_data = json.loads(result.output.strip('\n').split('\n')[-1])
            assert (
                result_data['data']['launchPipelineExecution']['__typename']
                == 'LaunchPipelineRunSuccess'
            )
        except Exception as e:
            raise Exception('Failed with {} Exception: {}'.format(result.output, e))


def test_start_execution_file():
    variables = seven.json.dumps(
        {
            'executionParams': {
                'selector': {
                    'pipelineName': 'math',
                    'repositoryLocationName': '<<in_process>>',
                    'repositoryName': 'test',
                },
                'runConfigData': {'solids': {'add_one': {'inputs': {'num': {'value': 123}}}}},
                'mode': 'default',
            }
        }
    )

    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')
    with dagster_cli_runner() as runner:
        result = runner.invoke(
            ui,
            [
                '-w',
                workspace_path,
                '-v',
                variables,
                '--file',
                file_relative_path(__file__, './execute.graphql'),
            ],
        )

        assert result.exit_code == 0
        result_data = json.loads(result.output.strip('\n').split('\n')[-1])
        assert (
            result_data['data']['launchPipelineExecution']['__typename']
            == 'LaunchPipelineRunSuccess'
        )


def test_start_execution_save_output():
    '''
    Test that the --output flag saves the GraphQL response to the specified file
    '''

    variables = seven.json.dumps(
        {
            'executionParams': {
                'selector': {
                    'repositoryLocationName': '<<in_process>>',
                    'repositoryName': 'test',
                    'pipelineName': 'math',
                },
                'runConfigData': {'solids': {'add_one': {'inputs': {'num': {'value': 123}}}}},
                'mode': 'default',
            }
        }
    )

    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')

    with dagster_cli_runner() as runner:
        with seven.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, 'output_file')

            result = runner.invoke(
                ui,
                [
                    '-w',
                    workspace_path,
                    '-v',
                    variables,
                    '--file',
                    file_relative_path(__file__, './execute.graphql'),
                    '--output',
                    file_name,
                ],
            )

            assert result.exit_code == 0

            assert os.path.isfile(file_name)
            with open(file_name, 'r') as f:
                lines = f.readlines()
                result_data = json.loads(lines[-1])
                assert (
                    result_data['data']['launchPipelineExecution']['__typename']
                    == 'LaunchPipelineRunSuccess'
                )


def test_start_execution_predefined():
    variables = seven.json.dumps(
        {
            'executionParams': {
                'selector': {
                    'repositoryLocationName': '<<in_process>>',
                    'repositoryName': 'test',
                    'pipelineName': 'math',
                },
                'runConfigData': {'solids': {'add_one': {'inputs': {'num': {'value': 123}}}}},
                'mode': 'default',
            }
        }
    )

    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')

    with dagster_cli_runner() as runner:
        result = runner.invoke(
            ui, ['-w', workspace_path, '-v', variables, '-p', 'launchPipelineExecution']
        )
        assert result.exit_code == 0
        result_data = json.loads(result.output.strip('\n').split('\n')[-1])
        if not result_data.get('data'):
            raise Exception(result_data)
        assert (
            result_data['data']['launchPipelineExecution']['__typename']
            == 'LaunchPipelineRunSuccess'
        )


def test_logs_in_start_execution_predefined():
    variables = seven.json.dumps(
        {
            'executionParams': {
                'selector': {
                    'repositoryLocationName': '<<in_process>>',
                    'repositoryName': 'test',
                    'pipelineName': 'math',
                },
                'runConfigData': {'solids': {'add_one': {'inputs': {'num': {'value': 123}}}}},
                'mode': 'default',
            }
        }
    )

    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance.local_temp(temp_dir)

        runner = CliRunner(env={'DAGSTER_HOME': temp_dir})
        result = runner.invoke(
            ui, ['-w', workspace_path, '-v', variables, '-p', 'launchPipelineExecution']
        )
        assert result.exit_code == 0
        result_data = json.loads(result.output.strip('\n').split('\n')[-1])
        assert (
            result_data['data']['launchPipelineExecution']['__typename']
            == 'LaunchPipelineRunSuccess'
        )
        run_id = result_data['data']['launchPipelineExecution']['run']['runId']

        # allow FS events to flush
        retries = 5
        while retries != 0 and not _is_done(instance, run_id):
            time.sleep(0.333)
            retries -= 1

        # assert that the watching run storage captured the run correctly from the other process
        run = instance.get_run_by_id(run_id)

        assert run.status == PipelineRunStatus.SUCCESS


def _is_done(instance, run_id):
    return instance.has_run(run_id) and instance.get_run_by_id(run_id).is_finished


def test_solid_selection():
    variables = seven.json.dumps(
        {
            'executionParams': {
                'mode': 'default',
                'runConfigData': {},
                'selector': {
                    'repositoryLocationName': '<<in_process>>',
                    'repositoryName': 'test',
                    'pipelineName': 'subset_test',
                    'solidSelection': ['no_config'],
                },
                'executionMetadata': {'runId': '1234'},
                'stepKeys': ['no_config.compute'],
            },
        }
    )
    workspace_path = file_relative_path(__file__, './cli_test_workspace.yaml')

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ['-w', workspace_path, '-v', variables, '-p', 'executePlan'])

        assert result.exit_code == 0

        try:
            result_data = json.loads(result.output.strip('\n').split('\n')[-1])
            assert result_data['data']['executePlan']['__typename'] == 'ExecutePlanSuccess'
        except Exception as e:
            raise Exception('Failed with {} Exception: {}'.format(result.output, e))



from dagster_graphql import dauphin

from dagster import check
from dagster.core.snap import PipelineSnapshot
from dagster.core.types.dagster_type import DagsterTypeKind

from .config_types import DauphinConfigType, to_dauphin_config_type


def config_type_for_schema(pipeline_snapshot, schema_key):
    return (
        to_dauphin_config_type(pipeline_snapshot.config_schema_snapshot, schema_key)
        if schema_key
        else None
    )


def to_dauphin_dagster_type(pipeline_snapshot, dagster_type_key):
    check.str_param(dagster_type_key, 'dagster_type_key')
    check.inst_param(pipeline_snapshot, pipeline_snapshot, PipelineSnapshot)

    dagster_type_meta = pipeline_snapshot.dagster_type_namespace_snapshot.get_dagster_type_snap(
        dagster_type_key
    )

    base_args = dict(
        key=dagster_type_meta.key,
        name=dagster_type_meta.name,
        display_name=dagster_type_meta.display_name,
        description=dagster_type_meta.description,
        is_builtin=dagster_type_meta.is_builtin,
        is_nullable=dagster_type_meta.kind == DagsterTypeKind.NULLABLE,
        is_list=dagster_type_meta.kind == DagsterTypeKind.LIST,
        is_nothing=dagster_type_meta.kind == DagsterTypeKind.NOTHING,
        input_schema_type=config_type_for_schema(
            pipeline_snapshot, dagster_type_meta.loader_schema_key,
        ),
        output_schema_type=config_type_for_schema(
            pipeline_snapshot, dagster_type_meta.materializer_schema_key,
        ),
        inner_types=list(
            map(
                lambda key: to_dauphin_dagster_type(pipeline_snapshot, key),
                dagster_type_meta.type_param_keys,
            )
        ),
    )

    if dagster_type_meta.kind == DagsterTypeKind.LIST:
        base_args['of_type'] = to_dauphin_dagster_type(
            pipeline_snapshot, dagster_type_meta.type_param_keys[0]
        )
        return DauphinListDagsterType(**base_args)
    elif dagster_type_meta.kind == DagsterTypeKind.NULLABLE:
        base_args['of_type'] = to_dauphin_dagster_type(
            pipeline_snapshot, dagster_type_meta.type_param_keys[0]
        )
        return DauphinNullableDagsterType(**base_args)
    else:
        return DauphinRegularDagsterType(**base_args)


class DauphinDagsterType(dauphin.Interface):
    class Meta(object):
        name = 'DagsterType'

    key = dauphin.NonNull(dauphin.String)
    name = dauphin.String()
    display_name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()

    is_nullable = dauphin.NonNull(dauphin.Boolean)
    is_list = dauphin.NonNull(dauphin.Boolean)
    is_builtin = dauphin.NonNull(dauphin.Boolean)
    is_nothing = dauphin.NonNull(dauphin.Boolean)

    input_schema_type = dauphin.Field(DauphinConfigType)
    output_schema_type = dauphin.Field(DauphinConfigType)

    inner_types = dauphin.non_null_list('DagsterType')


class DauphinRegularDagsterType(dauphin.ObjectType):
    class Meta(object):
        name = 'RegularDagsterType'
        interfaces = [DauphinDagsterType]


class DauphinWrappingDagsterType(dauphin.Interface):
    class Meta(object):
        name = 'WrappingDagsterType'

    of_type = dauphin.Field(dauphin.NonNull(DauphinDagsterType))


class DauphinListDagsterType(dauphin.ObjectType):
    class Meta(object):
        name = 'ListDagsterType'
        interfaces = [DauphinDagsterType, DauphinWrappingDagsterType]


class DauphinNullableDagsterType(dauphin.ObjectType):
    class Meta(object):
        name = 'NullableDagsterType'
        interfaces = [DauphinDagsterType, DauphinWrappingDagsterType]



from dagster_graphql import dauphin
from dagster_graphql.implementation.fetch_assets import get_asset_events, get_asset_run_ids
from dagster_graphql.implementation.fetch_runs import get_run_by_id
from dagster_graphql.schema.runs import construct_basic_params

from dagster import check
from dagster.core.events.log import EventRecord
from dagster.core.storage.pipeline_run import PipelineRunsFilter

from .errors import DauphinError


class DauphinAssetKey(dauphin.ObjectType):
    class Meta(object):
        name = 'AssetKey'

    path = dauphin.non_null_list(dauphin.String)


class DauphinAsset(dauphin.ObjectType):
    class Meta(object):
        name = 'Asset'

    key = dauphin.NonNull('AssetKey')
    assetMaterializations = dauphin.Field(
        dauphin.non_null_list('AssetMaterialization'), cursor=dauphin.String(), limit=dauphin.Int(),
    )
    runs = dauphin.Field(
        dauphin.non_null_list('PipelineRun'), cursor=dauphin.String(), limit=dauphin.Int(),
    )

    def resolve_assetMaterializations(self, graphene_info, **kwargs):
        return [
            graphene_info.schema.type_named('AssetMaterialization')(event=event)
            for event in get_asset_events(
                graphene_info, self.key, kwargs.get('cursor'), kwargs.get('limit')
            )
        ]

    def resolve_runs(self, graphene_info, **kwargs):
        cursor = kwargs.get('cursor')
        limit = kwargs.get('limit')

        run_ids = get_asset_run_ids(graphene_info, self.key)

        if not run_ids:
            return []

        # for now, handle cursor/limit here instead of in the DB layer
        if cursor:
            try:
                idx = run_ids.index(cursor)
                run_ids = run_ids[idx:]
            except ValueError:
                return []

        if limit:
            run_ids = run_ids[:limit]

        return [
            graphene_info.schema.type_named('PipelineRun')(r)
            for r in graphene_info.context.instance.get_runs(
                filters=PipelineRunsFilter(run_ids=run_ids)
            )
        ]


class DauphinAssetMaterialization(dauphin.ObjectType):
    class Meta(object):
        name = 'AssetMaterialization'

    def __init__(self, event):
        self._event = check.inst_param(event, 'event', EventRecord)

    materializationEvent = dauphin.NonNull('StepMaterializationEvent')
    runOrError = dauphin.NonNull('PipelineRunOrError')

    def resolve_materializationEvent(self, graphene_info):
        return graphene_info.schema.type_named('StepMaterializationEvent')(
            materialization=self._event.dagster_event.step_materialization_data.materialization,
            **construct_basic_params(self._event)
        )

    def resolve_runOrError(self, graphene_info):
        return get_run_by_id(graphene_info, self._event.run_id)


class DauphinAssetsNotSupportedError(dauphin.ObjectType):
    class Meta(object):
        name = 'AssetsNotSupportedError'
        interfaces = (DauphinError,)


class DauphinAssetsOrError(dauphin.Union):
    class Meta(object):
        name = 'AssetsOrError'
        types = ('AssetConnection', 'AssetsNotSupportedError')


class DauphinAssetConnection(dauphin.ObjectType):
    class Meta(object):
        name = 'AssetConnection'

    nodes = dauphin.non_null_list('Asset')


class DauphinAssetOrError(dauphin.Union):
    class Meta(object):
        name = 'AssetOrError'
        types = ('Asset', 'AssetsNotSupportedError')



from dagster.core.utils import check_dagster_package_version

from .dauphin_registry import DauphinRegistry
from .version import __version__

dauphin = DauphinRegistry()
check_dagster_package_version('dagster-graphql', __version__)



import pytest

from dagster import seven
from dagster.core.instance import DagsterInstance

from .setup import define_test_context


@pytest.yield_fixture(scope='function')
def graphql_context():
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance.local_temp(
            temp_dir,
            overrides={
                'scheduler': {
                    'module': 'dagster.utils.test',
                    'class': 'FilesystemTestScheduler',
                    'config': {'base_dir': temp_dir},
                }
            },
        )
        yield define_test_context(instance)



import re
from collections import OrderedDict

from dagster_graphql.test.preset_query import execute_preset_query

from .graphql_context_test_suite import ReadonlyGraphQLContextTestMatrix


class TestPresets(ReadonlyGraphQLContextTestMatrix):
    def test_basic_preset_query_no_presets(self, graphql_context):
        result = execute_preset_query('csv_hello_world_two', graphql_context)
        assert result.data == OrderedDict(
            [('pipelineOrError', OrderedDict([('name', 'csv_hello_world_two'), ('presets', [])]))]
        )

    def test_basic_preset_query_with_presets(self, graphql_context, snapshot):
        result = execute_preset_query('csv_hello_world', graphql_context)

        assert [
            preset_data['name'] for preset_data in result.data['pipelineOrError']['presets']
        ] == ['prod', 'test', 'test_inline',]

        # Remove local filepath from snapshot
        result.data['pipelineOrError']['presets'][2]['runConfigYaml'] = re.sub(
            r'num: .*/data/num.csv',
            'num: /data/num.csv',
            result.data['pipelineOrError']['presets'][2]['runConfigYaml'],
        )
        snapshot.assert_match(result.data)



from dagster_graphql.cli import execute_query

from dagster.core.errors import DagsterError
from dagster.core.events.log import DagsterEventRecord
from dagster.core.instance import DagsterInstance
from dagster.serdes import deserialize_json_to_dagster_namedtuple

from .query import EXECUTE_PLAN_MUTATION, RAW_EXECUTE_PLAN_MUTATION
from .util import HANDLED_EVENTS, dagster_event_from_dict


class DagsterGraphQLClientError(DagsterError):
    '''Indicates that some error has occurred when executing a GraphQL query against
    dagster-graphql'''


def execute_execute_plan_mutation(workspace, variables, instance_ref=None):
    instance = (
        DagsterInstance.from_ref(instance_ref) if instance_ref else DagsterInstance.ephemeral()
    )
    res = execute_query(
        workspace, EXECUTE_PLAN_MUTATION, variables, use_sync_executor=True, instance=instance
    )
    instance.dispose()
    handle_execution_errors(res, 'executePlan')
    return handle_execute_plan_result(res)


def handle_execution_errors(res, expected_type):
    if res is None:
        raise DagsterGraphQLClientError('Unhandled error type. Missing query response')

    if res.get('errors'):
        raise DagsterGraphQLClientError(
            'Internal error in GraphQL request. Response: {}'.format(res)
        )

    if not res.get('data', {}).get(expected_type, {}).get('__typename'):
        raise DagsterGraphQLClientError('Unexpected response type. Response: {}'.format(res))


def execute_execute_plan_mutation_raw(recon_repo, variables, instance_ref=None):
    '''The underlying mutation returns the DagsterEventRecords serialized as strings, rather
    than dict representations of the DagsterEvents, thus "raw". This method in turn returns a
    stream of DagsterEventRecords, not DagsterEvents.'''

    instance = (
        DagsterInstance.from_ref(instance_ref) if instance_ref else DagsterInstance.ephemeral()
    )
    res = execute_query(recon_repo, RAW_EXECUTE_PLAN_MUTATION, variables, instance=instance)
    handle_execution_errors(res, 'executePlan')
    return handle_execute_plan_result_raw(res)


def handle_error_states(res_type, res_data):
    if res_type == 'InvalidStepError':
        raise DagsterGraphQLClientError(
            'invalid step {step_key}'.format(step_key=res_data['invalidStepKey'])
        )

    if res_type == 'PipelineConfigValidationInvalid':
        errors = [err['message'] for err in res_data['errors']]
        raise DagsterGraphQLClientError(
            'Pipeline configuration invalid:\n{errors}'.format(errors='\n'.join(errors))
        )

    if res_type == 'PipelineNotFoundError':
        raise DagsterGraphQLClientError(
            'Pipeline "{pipeline_name}" not found: {message}:'.format(
                pipeline_name=res_data['pipelineName'], message=res_data['message']
            )
        )

    if res_type == 'PythonError':
        raise DagsterGraphQLClientError(
            'Subplan execution failed: {message}\n{stack}'.format(
                message=res_data['message'], stack=res_data['stack']
            )
        )


def handle_execute_plan_result(res):
    res_data = res['data']['executePlan']

    res_type = res_data['__typename']

    handle_error_states(res_type, res_data)

    if res_type == 'ExecutePlanSuccess':
        pipeline_name = res_data['pipeline']['name']

        return [
            dagster_event_from_dict(e, pipeline_name)
            for e in res_data['stepEvents']
            if e['__typename'] in HANDLED_EVENTS
        ]

    raise DagsterGraphQLClientError('Unexpected result type')


def handle_execute_plan_result_raw(res):
    res_data = res['data']['executePlan']

    res_type = res_data['__typename']

    handle_error_states(res_type, res_data)

    if res_type == 'ExecutePlanSuccess':
        raw_event_records = [
            DagsterEventRecord(
                event_record.error_info,
                event_record.message,
                event_record.level,
                event_record.user_message,
                event_record.run_id,
                event_record.timestamp,
                event_record.step_key,
                event_record.pipeline_name,
                event_record.dagster_event,
            )
            for event_record in [
                deserialize_json_to_dagster_namedtuple(e) for e in res_data['rawEventRecords']
            ]
        ]
        return raw_event_records

    raise DagsterGraphQLClientError('Unexpected result type')
